function Features = Feature_extraction_final(ecg, fs)
%% Signal Normalization
signal = (ecg-mean(ecg))./std(ecg);

Features = [];

%% Baseline removal
signal_baseline_free = remove_baseline_ecg(signal, fs);
baseline_signal = signal - signal_baseline_free;

Features_baseline_signal = array_characteristics(baseline_signal);
Features = [Features Features_baseline_signal bandpower(baseline_signal)];
%%
[~, R] = pan_tompkin_algorithm_segmented(signal_baseline_free, fs);
R = R(R > 0);
R_peaks = unique(R);

[Q_ind, R_ind, S_ind]  = get_qrs_complex(signal_baseline_free, R_peaks, fs);

%% Amplitudes:
Q_amplitudes = signal_baseline_free(Q_ind);
R_amplitudes = signal_baseline_free(R_ind);
S_amplitudes = signal_baseline_free(S_ind);

RS_amplitude = R_amplitudes - S_amplitudes;
RQ_amplitude = R_amplitudes - Q_amplitudes;
qrs_width = S_ind - Q_ind;

Features_RS_amplitude = array_characteristics(RS_amplitude);
Features_RQ_amplitude = array_characteristics(RQ_amplitude);
Features_qrs_width = array_characteristics(qrs_width);

Features = [Features Features_RS_amplitude Features_RQ_amplitude Features_qrs_width];

%% HR features
RR_intervals = diff(R_peaks);
HR = 60*fs./RR_intervals;
HR_diff = diff(HR);
HR_diff_2 = diff(HR_diff);
HR_diff_3 = diff(HR_diff_2);

Features_HR = array_characteristics(HR);
Features_HR_diff =  array_characteristics(HR_diff);
Features_HR_diff_2 =  array_characteristics(HR_diff_2);
Features_HR_diff_3 =  array_characteristics(HR_diff_3);

Features = [Features Features_HR Features_HR_diff Features_HR_diff_2 Features_HR_diff_3];

%% Bradycardia test Find QRS sequence of 3 members with HR < 40 bpm - Number of times where this happens is our feature

HR_limit_down = 40;
bradycardia_feature = numel(find(HR <= HR_limit_down))/numel(HR);
Features = [Features bradycardia_feature];
%% Extreme bradycardia feature - Heart rate lower than 40 bpm for 5 consecutive beats

belowThreshold = [0, HR < HR_limit_down, 0];
edges = diff(belowThreshold);
rising = find(edges == 1);     %rising/falling edges
falling = find(edges == -1);  
spanWidth = falling - rising; 
wideEnough = spanWidth >= 5;   %% 5 consecutive values

if (isempty(find(wideEnough)))
    feature_extreme_bradycardia = 0;
else
    feature_extreme_bradycardia = 1;
end

Features = [Features feature_extreme_bradycardia];

%% Tachycardia test Find QRS sequence of 12 beats with HR > 130 bpm - Number of times where this happens is our feature
HR_limit_up = 130;
tachycardia_feature = numel(find(HR >= HR_limit_up))/numel(HR);

Features = [Features tachycardia_feature];

%% Extreme Tachycardia - Heart rate higher than 140 bpm for 17 consecutive beats

HR_limit_up_extreme = 140;

aboveThreshold = [0, HR > HR_limit_up_extreme, 0];
edges = diff(aboveThreshold);
rising = find(edges == 1);     %rising/falling edges
falling = find(edges == -1);  
spanWidth = falling - rising; 
min_element = min([17, numel(R_peaks)]);
wideEnough = spanWidth >= min_element;   

if (isempty(find(wideEnough)))
    feature_extreme_tachycardia = 0;
else
    feature_extreme_tachycardia = 1;
end

Features = [Features feature_extreme_tachycardia];
%% 

HR_down_1 = 46;
HR_down_2 = 40;

HR_up_1 = 100;
HR_up_2 = 130;

if (mean(HR) < HR_down_1)
    feature_down_1 = 1;
else 
    feature_down_1 = 0;
end

if (mean(HR) < HR_down_2)
    feature_down_2 = 1;
else 
    feature_down_2 = 0;
end

if (mean(HR) > HR_up_1)
    feature_up_1 = 1;
else 
    feature_up_1 = 0;
end

if (mean(HR) > HR_up_2)
    feature_up_2 = 1;
else 
    feature_up_2 = 0;
end

Features = [Features feature_down_1 feature_down_2 feature_up_1 feature_up_2];


%% Features for P wave - Normal sinus rhythm  0.12 < PR interval < 0.20 sec   -- Go from Q

Samples_from_Q_point = 0.01 * fs; 
duration_p_wave = 0.15 * fs; %% 120 ms < PR interval < 200 ms
Signals_p_wave = [];
Signals_p_wave_diff = [];
Signals_p_wave_diff_2 = [];
Signals_p_wave_diff_3 = [];

Signals_p_wave_norm = [];
Signals_p_wave_diff_norm = [];
Signals_p_wave_diff_2_norm = [];
Signals_p_wave_diff_3_norm = [];

Signals_p_wave_diff_sum = [];
sign_of_p_wave = [];
p_wave_peaks  = [];
PR_interval = [];

for i = 1 : length(Q_ind)
    
    index_end = Q_ind(i) - Samples_from_Q_point;
    index_start = index_end - duration_p_wave;
    index_start = round(index_start);
    index_end = round(index_end);
    
    if (index_start <= 0)
        continue
    else
        sig_p_wave = signal_baseline_free(index_start:index_end);
        sig_p_wave_diff = diff(sig_p_wave);
        sig_p_wave_diff_2 = diff(sig_p_wave_diff);
        sig_p_wave_diff_3 = diff(sig_p_wave_diff_2);
        
        Signals_p_wave = [Signals_p_wave sig_p_wave(:)];
        Signals_p_wave_diff = [Signals_p_wave_diff sig_p_wave_diff(:)];
        Signals_p_wave_diff_2 = [Signals_p_wave_diff_2 sig_p_wave_diff_2(:)];
        Signals_p_wave_diff_3 = [Signals_p_wave_diff_3 sig_p_wave_diff_3(:)];  
        
        Signals_p_wave_norm = [Signals_p_wave_norm norm(sig_p_wave, 2)];
        Signals_p_wave_diff_norm = [Signals_p_wave_diff_norm norm(sig_p_wave_diff, 2)];
        Signals_p_wave_diff_2_norm = [Signals_p_wave_diff_2_norm norm(sig_p_wave_diff_2, 2)];
        Signals_p_wave_diff_3_norm = [Signals_p_wave_diff_3_norm norm(sig_p_wave_diff_3, 2)];
        
        Signals_p_wave_diff_sum = [Signals_p_wave_diff_sum sum(sig_p_wave_diff)];
        
        [~, p_peak] = max(sig_p_wave);
        ind_p_peak = p_peak + index_start - 1;
        p_wave_peaks = [p_wave_peaks ind_p_peak];
        PR_interval = [PR_interval (Q_ind(i) - ind_p_peak)];
        sign_of_p_wave = [sign_of_p_wave sign(max(sig_p_wave))];  %% For junctional arrhythmia, P wave can be negative
        
        
    end
    
end


Features = [Features sum(sign_of_p_wave)/length(Q_ind) std(sign_of_p_wave)/length(Q_ind)];


Feat_1 = array_characteristics(min(Signals_p_wave));
Feat_2 = array_characteristics(max(Signals_p_wave));
Feat_3 = array_characteristics(mean(Signals_p_wave));
Feat_4 = array_characteristics(median(Signals_p_wave));
Feat_5 = array_characteristics(std(Signals_p_wave));

Features = [Features Feat_1 Feat_2 Feat_3 Feat_4 Feat_5];

Feat_6 = array_characteristics(min(Signals_p_wave_diff));
Feat_7 = array_characteristics(max(Signals_p_wave_diff));
Feat_8 = array_characteristics(mean(Signals_p_wave_diff));
Feat_9 = array_characteristics(median(Signals_p_wave_diff));
Feat_10 = array_characteristics(std(Signals_p_wave_diff));

Features = [Features Feat_6 Feat_7 Feat_8 Feat_9 Feat_10];

Feat_11 = array_characteristics(min(Signals_p_wave_diff_2));
Feat_12 = array_characteristics(max(Signals_p_wave_diff_2));
Feat_13 = array_characteristics(mean(Signals_p_wave_diff_2));
Feat_14 = array_characteristics(median(Signals_p_wave_diff_2));
Feat_15 = array_characteristics(std(Signals_p_wave_diff_2));

Features = [Features Feat_11 Feat_12 Feat_13 Feat_14 Feat_15];

Feat_16 = array_characteristics(min(Signals_p_wave_diff_3));
Feat_17 = array_characteristics(max(Signals_p_wave_diff_3));
Feat_18 = array_characteristics(mean(Signals_p_wave_diff_3));
Feat_19 = array_characteristics(median(Signals_p_wave_diff_3));
Feat_20 = array_characteristics(std(Signals_p_wave_diff_3));


Features = [Features Feat_16 Feat_17 Feat_18 Feat_19 Feat_20];

Feat_21 = array_characteristics(Signals_p_wave_norm);
Feat_22 = array_characteristics(Signals_p_wave_diff_norm);
Feat_23 = array_characteristics(Signals_p_wave_diff_2_norm);
Feat_24 = array_characteristics(Signals_p_wave_diff_3_norm);
Feat_25 = array_characteristics(Signals_p_wave_diff_sum);

Features = [Features Feat_21 Feat_22 Feat_23 Feat_24 Feat_25];



%% PR interval -  %% 120 ms < PR interval < 200 ms 
Features = [Features std(PR_interval)/std(RR_intervals)];
Features = [Features array_characteristics(PR_interval)];

%% PP interval 
PP_interval = diff(p_wave_peaks);

ratio_PP_RR_mean = mean(RR_intervals)/mean(PP_interval);
ratio_PP_RR_std = std(RR_intervals)/std(PP_interval);

Features = [Features ratio_PP_RR_mean ratio_PP_RR_std];
    
%% Premature ventricular or atrial contractions
R_ampl = ecg(R_peaks)./median(ecg(R_peaks));
percsmallPCs = sum(R_ampl<0.8)/length(R_ampl);
perchighPCs = sum(R_ampl>1.2)/length(R_ampl);

Features = [Features percsmallPCs perchighPCs];

%%  Normal sinus rhythm QT interval 0.35-0.43 secs
[~,~,~,delbeats, ~] = delineation_leadI_II(signal_baseline_free, fs);

QT_interval = (delbeats(:,8) - delbeats(:,4));

Features = [Features array_characteristics(QT_interval)];

%% HR group changes
num_beats = 5; 

HR_group = [];
Std_group = [];

for i = 1 : fix(length(HR)/num_beats)
    HR_group = HR(1 + (i-1)*num_beats : i*num_beats);
    Std_group = [Std_group std(HR_group)];
    
end

Features = [Features mean(Std_group) std(Std_group)];

%% Features for T wave - Ratio Rpeak/Twave

[~, ~, ~, delbeats, ~] = delineation_leadI_II(signal_baseline_free, fs);

Peak_T_wave = signal_baseline_free(delbeats(:,8));
Peak_R_wave =  signal_baseline_free(delbeats(:,5));

Ratio_R_T_amplitudes = Peak_R_wave./Peak_T_wave;

percent_t_above_r = numel(find(Ratio_R_T_amplitudes < 1))/numel(Peak_R_wave);
Features = [Features percent_t_above_r];
%% T wave similarities

q = 1;
n_right = 120;
offset = 20;

for n = 1 : length(S_ind)
    
    start_ind = S_ind(n) + offset;
    end_ind = S_ind(n) + n_right - 1;  
    
    if (start_ind <= 0 || end_ind <= 0)
        continue
    else
        if (end_ind >= length(signal_baseline_free) || start_ind >= length(signal_baseline_free))
            continue 
        else 
            pens_after(q,:) = signal_baseline_free(start_ind:end_ind);
            q = q + 1;
        end
    end
    
end

Average_t_wave = pens_after';

Features = [Features shanon_entropy(Average_t_wave(:))];

corr_matrix_t_wave = corr(Average_t_wave);

num_considered_t_waves = size(corr_matrix_t_wave, 1);

feature_corr_t_wave = (sum(corr_matrix_t_wave(:)) - num_considered_t_waves)/(num_considered_t_waves.^2 - num_considered_t_waves);

Features = [Features feature_corr_t_wave];

%% Concatenated T wave power features

T_wave_signal = Average_t_wave(:);
total_power_t_wave = bandpower(T_wave_signal);
f_low_t_wave = bandpower(T_wave_signal, fs, [0 5])/total_power_t_wave;
f_high_t_wave = bandpower(T_wave_signal, fs, [5 30])/total_power_t_wave;

Features = [Features f_low_t_wave f_high_t_wave];

%% Features for P wave

q = 1;
n_left = 140;
offset = 20;

for n = 1 : length(Q_ind)
    
    end_ind = Q_ind(n) - offset;
    start_ind = end_ind - n_left; 
    
    if (start_ind <= 0 || end_ind <= 0)
        continue
    else
        if (end_ind >= length(ecg) || start_ind >= length(ecg))
            continue 
        else 
            pens_before(q,:) = signal_baseline_free(start_ind:end_ind);
            q = q + 1;
        end
    end
    
end

Average_p_wave = pens_before';

Features = [Features shanon_entropy(Average_p_wave(:))];

corr_matrix_p_wave = corr(Average_p_wave);

num_considered_p_waves = size(corr_matrix_p_wave, 1);

feature_corr_p_wave = (sum(corr_matrix_p_wave(:)) - num_considered_p_waves)/(num_considered_p_waves.^2 - num_considered_p_waves);

Features = [Features feature_corr_p_wave];

%%  Concatenated P wave power features 

P_wave_signal = Average_p_wave(:);

total_power_p_wave = bandpower(P_wave_signal);

f_low_p_wave = bandpower(P_wave_signal, fs, [0 30])/total_power_p_wave;

Features = [Features f_low_p_wave];

%% Challenge features

[QRS,~,~] = qrs_detect2(ecg',0.25,0.6,fs);

if (length(QRS)<6)    
    feature_ch = 0;   
else
    RR_ch=diff(QRS')/fs;
    AFEv = comput_AFEv(RR_ch);
    if (AFEv > 1)
       feature_ch = 1;
    else
       feature_ch = 0;   
    end
    
end

Features = [Features feature_ch];
%% Power features ECG signal

output = [];
rrdiff = RR_intervals;

[qrsfid, sign_variable, ~] = qrs_detect2(ecg);
qrsdiff = diff(qrsfid);
output = [output sum(sign_variable)];

[waveletfid] = delineation_r_peaks_wavelet(ecg,fs, 2);
waveletdiff = diff(waveletfid);

[fastfid] = fast_delineation(ecg, fs);
fastdiff = diff(fastfid);
output=[output length(waveletdiff)/length(rrdiff) length(fastdiff)/length(rrdiff) length(qrsdiff)/length(rrdiff)];

tmp=length(removeoutliers(rrdiff))/length(rrdiff);    
output = [output tmp];

tmp = length(rrdiff)/length(ecg);
output = [output tmp];

tmp = bandpower(ecg, fs, [0.005 0.02])/bandpower(ecg);
output=[output tmp];

tmp = bandpower(ecg, fs, [0.02 0.05])/bandpower(ecg);
output=[output tmp];

tmp = bandpower(ecg, fs, [0.05 40])/bandpower(ecg);
output=[output tmp];

tmp = bandpower(ecg, fs, [40 50])/bandpower(ecg);
output=[output tmp];

tmp = bandpower(ecg, fs, [50 100])/bandpower(ecg);
output=[output tmp];

tmp = bandpower(ecg, fs, [100 fs/2 - 10])/bandpower(ecg);
output=[output tmp];

[tmp1, tmp2] = snr(ecg);
output=[output tmp1 tmp2]; 

Features = [Features output];
end

